;;;; -*- Scheme -*-
;;
;;  Copyright (c) 1997 Massachusetts Institute of Technology
;;  
;;  This material was developed by the Amorphous Computing project at the
;;  Massachusetts Institute of Technology Artificial Intelligence
;;  Laboratory.  Permission to copy this software, to redistribute it, and
;;  to use it for any purpose is granted, subject to the following
;;  restrictions and understandings.
;;  
;;  1. Any copy made of this software must include this copyright notice
;;  in full.
;;  
;;  2. Users of this software agree to make their best efforts (a) to
;;  return to the MIT Amorphous Computing project any improvements or
;;  extensions that they make, so that these may be included in future
;;  releases; and (b) to inform MIT of noteworthy uses of this software.
;;  
;;  3. All materials developed as a consequence of the use of this
;;  software shall duly acknowledge such use, in accordance with the usual
;;  standards of acknowledging credit in academic research.
;;  
;;  4. MIT has made no warrantee or representation that the operation of
;;  this software will be error-free, and MIT is under no obligation to
;;  provide any services, by way of maintenance, update, or otherwise.
;;  
;;  5. In conjunction with products arising from the use of this material,
;;  there shall be no use of the name of the Massachusetts Institute of
;;  Technology nor of any adaptation thereof in any advertising,
;;  promotional, or sales literature without prior written consent from
;;  MIT in each case.

(declare (usual-integrations))


;; Devices are essentially a procedure which performs some action or
;; makes some sensor reading.
;;
;; A device has an initialization procedure so that it may be
;; re-initialized to re-run a simulation.  When the initializer is
;; called, the simulation has an agenda (if one is used).

(define-structure
    (io-device
     (conc-name io-device.)
     (constructor make-io-device (initializer action)))

  (initializer %io-device-null-initializer   read-only true)
  (action      %io-device-null-action        read-only true))


;; Called by HLSIM to initialize the device at the beginning of a new
;; run of a simulation:

(define (io-device.initialize! dev sim)
  ((io-device.initializer dev) dev sim))


;; This is called from init.scm by the READ-SENSOR and APPLY-ACTUATOR
;; procedures:

(define (%simulation.io-device-operation sim index time device-name args)
  (let ((device (simulation.find-io-device sim device-name)))
    (apply (io-device.action device) device sim index time args)))




(define (%io-device-null-initializer dev sim)
  dev sim ; ignored
  unspecific)

(define (%io-device-null-action dev sim . args)
  dev sim args ; ignored
  #F)


;; Example: static projected intensity from a PNM image.
;;
;; Setup:
;;  (simulation.add-sensor! sim 'LIGHT-SENSOR
;;     (make-static-projected-pnm-intensity-sensor "picture.ppm"))
;; Gunk code:
;;  (read-sensor 'LIGHT-SENSOR)

(define (make-static-projected-PNM-colour-sensor image-file)
  (let ((pnm #F))

    (define (initializer dev sim)
      dev sim				; ignored
      (set! pnm (read-pnm image-file)))

    (define (action dev sim index time)
      dev time				; ignored
      (let ((x (vector-ref (simulation.x sim) index))
	    (y (vector-ref (simulation.y sim) index)))
	(if (or (<= x 0) (<= y 0) (>= x 1) (>= y 1))
	    0.0
	    (let ((px (truncate->exact (* x (pnm.width pnm))))
		  (py (truncate->exact (* (- 1 y) (pnm.height pnm)))))
	      (pnm.get-rgb pnm px py)))))
	    
    (make-io-device initializer action)))

(define (make-static-projected-PNM-intensity-sensor image-file)
  (let ((pnm #F))

    (define (initializer dev sim)
      dev sim				; ignored
      (set! pnm (read-pnm image-file)))

    (define (action dev sim index time)
      dev time				; ignored
      (let ((x (vector-ref (simulation.x sim) index))
	    (y (vector-ref (simulation.y sim) index)))
	(if (or (<= x 0) (<= y 0) (>= x 1) (>= y 1))
	    0.0
	    (let ((px (truncate->exact (* x (pnm.width pnm))))
		  (py (truncate->exact (* (- 1 y) (pnm.height pnm)))))
	      (pnm.get-brightness pnm px py)))))
	    
    (make-io-device initializer action)))
