;;;; -*- Scheme -*-
;;
;;  $Id: init.scm,v 1.11 1997/09/25 15:52:26 adams Exp $
;;
;;  Copyright (c) 1997 Massachusetts Institute of Technology
;;  
;;  This material was developed by the Amorphous Computing project at the
;;  Massachusetts Institute of Technology Artificial Intelligence
;;  Laboratory.  Permission to copy this software, to redistribute it, and
;;  to use it for any purpose is granted, subject to the following
;;  restrictions and understandings.
;;  
;;  1. Any copy made of this software must include this copyright notice
;;  in full.
;;  
;;  2. Users of this software agree to make their best efforts (a) to
;;  return to the MIT Amorphous Computing project any improvements or
;;  extensions that they make, so that these may be included in future
;;  releases; and (b) to inform MIT of noteworthy uses of this software.
;;  
;;  3. All materials developed as a consequence of the use of this
;;  software shall duly acknowledge such use, in accordance with the usual
;;  standards of acknowledging credit in academic research.
;;  
;;  4. MIT has made no warrantee or representation that the operation of
;;  this software will be error-free, and MIT is under no obligation to
;;  provide any services, by way of maintenance, update, or otherwise.
;;  
;;  5. In conjunction with products arising from the use of this material,
;;  there shall be no use of the name of the Massachusetts Institute of
;;  Technology nor of any adaptation thereof in any advertising,
;;  promotional, or sales literature without prior written consent from
;;  MIT in each case.

;; INIT
;;
;; Initialization code for the processor.  All procedures should be
;; written in hand-CPS converted form.
;; This file should always be loaded into the simulator first, e.g.
;;  (simulator.load sim "init" "foo" ...)

(declare (usual-integrations))
(declare (integrate-external "sim"))

;;  (call-with-current-continuation procedure) - standard procedure

(define (call-with-current-continuation cont fuel store procedure)
  (procedure cont
	     (fix:- fuel 1)
	     store
	     (lambda (cont* fuel store value)
	       cont*			; ignored
	       (cont value (fix:- fuel 1) store))))


;; (apply procedure args ... arg-list) - standard procedure

(define (apply cont fuel store procedure . args)
  ((access apply system-global-environment)
   (access apply system-global-environment)
   procedure cont fuel store args))

;;
;; (clock).  Returns local `cycle' count.
;;
(define (clock cont fuel store)
  (cont (- (vector-ref store store:counter) fuel) fuel store))

;;
;; (broadcast message).  Send message to all neighbours.
;; (carrier-sense?).     Is a message currently being transmitted nearby?
;;
(define (broadcast cont fuel store message)
  (%broadcast cont fuel store message))

(define (carrier-sense? cont fuel store)
  (cont (processor.carrier-sense? (store.processor store))
	fuel store))

;;
;; (processor-number).  Cheat.  My id within the simulation.
;; (the-simulation).  Cheat.
;;
(define (processor-number cont fuel store)
  (cont (store.index store) fuel store))

(define (the-simulation cont fuel store)
  (cont (store.simulation store) fuel store))

(define (color-me cont fuel store color)
  (cont (%color-me (store.simulation store)
		   (store.index store)
		   color)
	(fix:+ fuel 4)			; `cancel' the cost of calling COLOR-ME
	store))

;;
;; (externally thunk).  Call thunk at no cost.  Should be used to display
;; debugging output etc.
;;
(define (externally cont fuel store thunk)
  (thunk (lambda (value fuel* store)
	   fuel*			; ignored
	   (cont value
		 (fix:+ fuel 4)		; counter the cost of calling EXTERNALLY
		 store))
	 #x1ffffff			; lots of fuel (still a fixnum)
	 store))


;; Support for SENSORS and ACTUATORS
;;
;; (read-sensor sensor).
;; Simulates taking a sensor reading at my coordinates
;; sensor is a procedure defined externally that takes x,y,z coordinates.
;;
;; (apply-actuator actuator . args). Simulates actuation at my coordinates.
;; actuator is a procedure defined externally that takes x,y,z and args.

(define (io-device-operation cont fuel store device-name . args)
  (let ((sim (store.simulation store))
	(index (store.index store)))
    (cont (%simulation.io-device-operation
	   sim 
	   index
	   (- (vector-ref store store:counter) fuel) ; i.e. time
	   device-name
	   args)
	  (fix:- fuel 1)
	  store)))

(define read-sensor    io-device-operation)
(define apply-actuator io-device-operation)


;;
;; Thread support
;;  (make-thread thunk)

(define (current-thread cont fuel store)
  (cont (vector-ref store store:thread) (fix:- fuel 1) store))

(define (make-thread cont fuel store thunk)
  (let ((thread (%make-thread (store.processor store))))
    (set-thread.context-method! thread resume/0)
    (set-thread.context/1! thread thunk)
    (set-thread.context/2! thread %continuation:thread-finished)
    (cont thread (fix:- fuel 10) store)))

(define (start-thread cont fuel store thread)
  (if (not (thread.new? thread))
      (error:wrong-type-argument thread "new thread" 'start-thread))
  (thread.make-runnable! thread)
  (cont thread (fix:- fuel 1) store))

(define (%continuation:thread-finished value fuel store)
  (thread.kill! (vector-ref store store:thread) value)
  fuel)

(define (make&start-thread cont fuel store thunk)
  (make-thread (named-lambda (.cont thread fuel store)
		 (start-thread cont fuel store thread))
	       fuel
	       store
	       thunk))

(define (wait-multiple cont fuel store . events)
  (%select cont fuel store (list->vector events)))

(define (wait cont fuel store event)
  (%select (named-lambda (.cont value fuel store)
	     (cont (cdr value) fuel store))
	   fuel
	   store
	   (vector event)))

(define (make-event cont fuel store)
  (%make-event cont fuel store))

(define (make-timeout-event cont fuel store timeout-delay)
  (%make-timeout-event cont fuel store timeout-delay))

(define test-proc
  (let ((count 0))
    (lambda (cont fuel store)
      (set! count (+ count 1))
      (cont count (fix:- fuel 1) store))))


#|
(define (make-serializer)
  (let ((mutex (make-thread-mutex)))
    (define (serialized f)
      (define (serialized-f . args)
	(with-thread-mutex-locked mutex
				  (lambda ()
				    (apply f args))))
      serialized-f)
    serialized))
|#

(define (make-serializer cont fuel store)

  (define exit-event (%%make-event))
  (define owning-thread #F)		; current occupant

  (define (serialized cont fuel store procedure)

    (define (serialized-procedure cont fuel store . args)

      (if (eq? (vector-ref store store:thread) owning-thread) ; re-entrant?
	  ((access apply system-global-environment)
	   procedure cont fuel store args)

	  (let ((my-exit-event (%%make-event)))
	    ;; LET ... SET! is an atimic read and write
	    (let ((pending-exit-event exit-event)) ; atomic read ...
	      (set! exit-event my-exit-event) ; ... and write

	      (wait (lambda (value fuel store)
		      value		; ignored
		      (set! owning-thread (vector-ref store store:thread))
		      ((access apply system-global-environment)
		       procedure
		       (lambda (value fuel store)
			 (set! owning-thread #F)
			 (event.signal! my-exit-event 'done)
			 (cont value fuel store))
		       fuel store args))
		    fuel store pending-exit-event)))))
      
    (cont serialized-procedure fuel store))

  (event.signal! exit-event 'done)
  (cont serialized fuel store))


;; File initialization

(lambda (cont fuel store)
  (global-set! store '#(message-interrupt-handler #F #F) #F)
  ;(global-set! store '#(clock #F #F) clock)
  ;(global-set! store '#(broadcast #F #F) broadcast)
  ;(global-set! store '#(processor-number #F #F) processor-number)
  ;(global-set! store '#(the-simulation #F #F) the-simulation)
  ;(global-set! store '#(color-me #F #F) color-me)
  ;(global-set! store '#(externally #F #F) externally)
  ;(global-set! store '#(make-thread #F #F) make-thread)
  ;(global-set! store '#(current-thread #F #F) current-thread)
  ;(global-set! store '#(make-timeout-event #F #F) make-timeout-event)
  (cont unspecific fuel store))

