;;;; -*- Scheme -*-
;;
;;  $Id: neighbours.scm,v 1.8 1997/09/22 15:20:05 adams Exp $
;;
;;  Copyright (c) 1997 Massachusetts Institute of Technology
;;  
;;  This material was developed by the Amorphous Computing project at the
;;  Massachusetts Institute of Technology Artificial Intelligence
;;  Laboratory.  Permission to copy this software, to redistribute it, and
;;  to use it for any purpose is granted, subject to the following
;;  restrictions and understandings.
;;  
;;  1. Any copy made of this software must include this copyright notice
;;  in full.
;;  
;;  2. Users of this software agree to make their best efforts (a) to
;;  return to the MIT Amorphous Computing project any improvements or
;;  extensions that they make, so that these may be included in future
;;  releases; and (b) to inform MIT of noteworthy uses of this software.
;;  
;;  3. All materials developed as a consequence of the use of this
;;  software shall duly acknowledge such use, in accordance with the usual
;;  standards of acknowledging credit in academic research.
;;  
;;  4. MIT has made no warrantee or representation that the operation of
;;  this software will be error-free, and MIT is under no obligation to
;;  provide any services, by way of maintenance, update, or otherwise.
;;  
;;  5. In conjunction with products arising from the use of this material,
;;  there shall be no use of the name of the Massachusetts Institute of
;;  Technology nor of any adaptation thereof in any advertising,
;;  promotional, or sales literature without prior written consent from
;;  MIT in each case.

(declare (usual-integrations))


;;  Utilities for calculating communication graph.

;; POINT-NEIGHBOURS and CYCLIC-POINT-NEIGHBOURS take a set of points
;; (x(i),y(i),z(i)) and compute the neighbours within a radius
;; DIST-THRESHOLD.  CYCLIC-POINT-NEIGHBOURS assumes that the boundary
;; conditions are cyclic on a unit cube.  POINT-NEIGHBOURS had no
;; restriction on the values.
;;
;; Both algorithms work by dividing the space into a `grid' of buckets
;; in the (x,y) dimensions such that points within the desired
;; distance must be in the same or and adjacent 8-connected bucket.
;; CYCLIC-POINT-NEIGHBOURS has to stretch the buckets to fit the
;; cyclic boundarty conditions.
;;
;; The `%' versions of the above do the actual neighbour computation,
;; and are also used to generate data sets for the ASYMMETRIC-
;; versions of the above.


(define (point-neighbours x y z dist-threshold)
  (finish-point-neighbours (%point-neighbours x y z dist-threshold)))

(define (cyclic-point-neighbours x y z dist-threshold)
  (finish-point-neighbours (%cyclic-point-neighbours x y z dist-threshold)))


(define (%point-neighbours x y z dist-threshold)
  ;; X Y Z are vectors of flonums.
  ;; Returns a vector of vectors of indexes.

  (check-point-neighbours-args x y z dist-threshold 'POINT-NEIGHBOURS)
  (let* ((n (vector-length x))
	 (neighbours (make-vector n '())))

    (define (particle-add-neighbour! p1 p2 dist)
      dist
      (vector-set! neighbours p1 (cons p2 (vector-ref neighbours p1))))

    (define (vector-min v)
      (let loop ((value (vector-ref v 0)) (i 1))
	(if (= i (vector-length v))
	    value
	    (loop (min value (vector-ref v i)) (+ i 1)))))

    (define (vector-max v)
      (let loop ((value (vector-ref v 0)) (i 1))
	(if (= i (vector-length v))
	    value
	    (loop (max value (vector-ref v i)) (+ i 1)))))

    (define-integrable (d1 x1 x2)
      (flo:abs (flo:- x1 x2)))

    (define (dist^2 p1 p2)
      (define-integrable (sq x) (flo:* x x))
      (flo:+ (sq (d1 (vector-ref x p1) (vector-ref x p2)))
	     (flo:+ (sq (d1 (vector-ref y p1) (vector-ref y p2)))
		    (sq (d1 (vector-ref z p1) (vector-ref z p2))))))

    (let ((dist^2-threshold (* dist-threshold dist-threshold))
	  (min-x (vector-min x))
	  (max-x (vector-max x))
	  (min-y (vector-min y))
	  (max-y (vector-max y)))
      (let ((grid-dist-threshold	; guard against too small DIST-THRESHOLD
	     (max dist-threshold
		  (/ (max (- max-x min-x) (- max-y min-y)) (sqrt n)))))

	(define (ix x) (truncate->exact (/ (- x min-x) grid-dist-threshold)))
	(define (iy y) (truncate->exact (/ (- y min-y) grid-dist-threshold)))

	(let* ((n-xbuckets (+ (ix max-x) 1))
	       (n-ybuckets (+ (iy max-y) 1))
	       (grid       (build-grid n-xbuckets n-ybuckets ix iy x y)))
	  (define (bucket ix iy)
	    (vector-ref (vector-ref grid ix) iy))
	  (do ((ix 0 (fix:+ ix 1)))
	      ((fix:= ix n-xbuckets))
	    (do ((iy 0 (fix:+ iy 1)))
		((fix:= iy n-ybuckets))
	      (let ()
		(define (check-bucket p1 ix iy)
		  (define (check-point p2)
		    (if (fix:< p1 p2)
			(let ((dist^2 (dist^2 p1 p2)))
			  (if (flo:< dist^2 dist^2-threshold)
			      (begin
				(particle-add-neighbour! p1 p2 dist^2)
				(particle-add-neighbour! p2 p1 dist^2))))))
		  (if (and (fix:<= 0 ix) (fix:< ix n-xbuckets) ; grid boundary?
			   (fix:<= 0 iy) (fix:< iy n-ybuckets))
		      (for-each check-point (bucket ix iy))))
		  
		(for-each
		    (lambda (p1)
		      (check-bucket p1 ix iy)
		      (check-bucket p1 (fix:- ix 1) iy)
		      (check-bucket p1 (fix:+ ix 1) iy)
		      (check-bucket p1 ix (fix:- iy 1))
		      (check-bucket p1 ix (fix:+ iy 1))
		      (check-bucket p1 (fix:- ix 1) (fix:- iy 1))
		      (check-bucket p1 (fix:- ix 1) (fix:+ iy 1))
		      (check-bucket p1 (fix:+ ix 1) (fix:- iy 1))
		      (check-bucket p1 (fix:+ ix 1) (fix:+ iy 1)))
		  (bucket ix iy))))))))

    neighbours))


(define (%cyclic-point-neighbours x y z dist-threshold)
  ;; Connect particles to neighbours with cyclic boundary conditions.
  ;; X Y Z are vectors of flonums in the unit cube.
  ;; Returns a vector of vectors of indexes.

  (check-point-neighbours-args x y z dist-threshold 'POINT-NEIGHBOURS)
  (let* ((n (vector-length x))
	 (neighbours (make-vector n '())))

    (define (particle-add-neighbour! p1 p2 dist)
      dist
      (vector-set! neighbours p1 (cons p2 (vector-ref neighbours p1))))

    ;; Distance between two points.  This formulation contains no
    ;; conditionals and so open-codes into straight line code with
    ;; maximal CSE.
    (define-integrable (d1 x1 x2 half)
      (flo:- half (flo:abs (flo:- (flo:abs (flo:- x1 x2)) half))))

    (define (dist^2 p1 p2)
      (define-integrable (sq x) (flo:* x x))
      (flo:+ (sq (d1 (vector-ref x p1) (vector-ref x p2) 0.5))
	     (flo:+ (sq (d1 (vector-ref y p1) (vector-ref y p2) 0.5))
		    (sq (d1 (vector-ref z p1) (vector-ref z p2) 0.5)))))

    (let ((dist^2-threshold (* dist-threshold dist-threshold)))
    
      (let* ((adjusted-dist-threshold (max dist-threshold (expt n -.5)))
	     ;; ADJUSTED-DIST-THRESHOLD prevents us creating too many
	     ;; buckets when DIST-THRESHOLD is small.
	     (n-buckets
	      (max 1 (- (truncate->exact (/ 1.0 adjusted-dist-threshold)) 1)))
	     (bucket-width (/ 1.0 n-buckets)))
	(define (ix x) (truncate->exact (/ x bucket-width)))
	(define (iy y) (truncate->exact (/ y bucket-width)))
	(let ((grid (build-grid n-buckets n-buckets ix iy x y)))
	  (define (bucket ix iy)
	    (vector-ref (vector-ref grid (modulo ix n-buckets))
			(modulo iy n-buckets)))
	  (do ((ix 0 (fix:+ ix 1)))
	      ((fix:= ix n-buckets))
	    (do ((iy 0 (fix:+ iy 1)))
		((fix:= iy n-buckets))
	      (let ()
		(define (check-bucket p1 ix iy)
		  (for-each
		      (lambda (p2)
			(if (fix:< p1 p2)
			    (let ((dist^2 (dist^2 p1 p2)))
			      (if (flo:< dist^2 dist^2-threshold)
				  (begin
				    (particle-add-neighbour! p1 p2 dist^2)
				    (particle-add-neighbour! p2 p1 dist^2))))))
		    (bucket ix iy)))
		  
		(for-each
		    (lambda (p1)
		      (check-bucket p1 ix iy)
		      ;; The following checks prevent double-counting
		      ;; in the cases where the grid is smalled than 3x3.
		      (if (fix:> n-buckets 1) 
			  (begin
			    (check-bucket p1 (fix:+ ix 1) iy)
			    (check-bucket p1 ix (fix:+ iy 1))
			    (check-bucket p1 (fix:+ ix 1) (fix:+ iy 1))
			    (if (fix:> n-buckets 2)
				(begin
				  (check-bucket p1 ix (fix:- iy 1))
				  (check-bucket p1 (fix:- ix 1) iy)
				  (check-bucket p1 (fix:- ix 1) (fix:- iy 1))
				  (check-bucket p1 (fix:- ix 1) (fix:+ iy 1))
				  (check-bucket p1 (fix:+ ix 1) (fix:- iy 1)))))))
		  (bucket ix iy))))))))

    neighbours))


(define (check-point-neighbours-args x y z dist-threshold procedure)
  (define (check-v v)
    (guarantee-vector v procedure)
    (do ((i 0 (fix:+ i 1)))
	((fix:= i (vector-length v)))
      (if (not (flo:flonum? (vector-ref v i)))
	  (error:wrong-type-argument v "vector of flonums" procedure))))
  (check-v x)
  (check-v y)
  (check-v z)
  (if (not (= (vector-length x) (vector-length y)))
      (error:wrong-type-argument y "vector of required length" procedure))
  (if (not (= (vector-length x) (vector-length z)))
      (error:wrong-type-argument y "vector of required length" procedure))
  (if (not (flo:flonum? dist-threshold))
      (error:wrong-type-argument dist-threshold "flonum" procedure)))


(define (build-grid nx ny fx fy x y)
  ;; Build a NX by NY grid of point indexes
  (let ((grid
	 (make-initialized-vector nx
	   (lambda (i) i (make-vector ny '())))))
    (do ((i 0 (fix:+ i 1)))
	((fix:= i (vector-length x)))
      (let  ((ix (fx (vector-ref x i)))
	     (iy (fy (vector-ref y i))))
	(vector-set! (vector-ref grid ix) iy
		     (cons i (vector-ref (vector-ref grid ix) iy)))))
    grid))


(define (finish-point-neighbours neighbours)
  (let ((neighbours (vector-map neighbours list->vector))
	(n (vector-length neighbours)))
    (let ((counts (vector->list (vector-map neighbours vector-length))))
      (let ((*min  (reduce min 1e9 counts))
	    (*max  (reduce max 0 counts))
	    (*mean (exact->inexact (/ (reduce + 0 counts) n))))
	(newline)
	(display "; ")
	(display n)
	(display " processors, neighbours: ")
	(display *min) (display " min, ")
	(display *mean) (display " mean, ")
	(display *max) (display " max.")))
    neighbours))


(define (asymmetric-generator symmetric-generator)
  (define (adjust inner-list outer-list)
    (list-transform-positive outer-list
      (lambda (e)
	(or (memq e inner-list)
	    (> (random 1.) 0.5)))))
  (lambda (x y z r-inner r-outer)
    (let ((inner-neighbours (symmetric-generator x y z r-inner))
	  (outer-neighbours (symmetric-generator x y z r-outer)))
      ;; Both INNER-NEIGHBOURS and OUTER-NEIGHBOURS are vectors of lists
      (let ((n (vector-length inner-neighbours)))
	(do ((i 0 (+ i 1)))
	    ((= i n))
	  (vector-set! inner-neighbours i
		       (adjust (vector-ref inner-neighbours i)
			       (vector-ref outer-neighbours i)))))
      (finish-point-neighbours inner-neighbours))))
      

(define asymmetric-point-neighbours
  (asymmetric-generator %point-neighbours))

(define asymmetric-cyclic-point-neighbours
  (asymmetric-generator %cyclic-point-neighbours))


(define (random-2d-points n)
  (values
   (make-initialized-vector n (lambda (i) i (random 1.0)))
   (make-initialized-vector n (lambda (i) i (random 1.0)))))

(define (pseudo-uniform-2d-points n #!optional per-square)
  (let ((per-square (if (default-object? per-square) 1 per-square)))
    (let ((side (truncate->exact (sqrt (/ n per-square)))))
      (let ((rows side)
	    (cols side))
	(let ((xs (make-vector n 0.0))
	      (ys (make-vector n 0.0))
	      (slots (make-initialized-vector n identity-procedure))
	      (i 0))
	  (define (make x y)
	    (let ((si (+ i (random (- n i)))))
	      (let ((s (vector-ref slots si)))
		(vector-set! slots si (vector-ref slots i))
		(vector-set! xs s x)
		(vector-set! ys s y)
		(set! i (+ i 1)))))

	  (do ((i 0 (+ i 1)))
	      ((= i cols))
	    (do ((j 0 (+ j 1)))
		((= j rows))
	      (do ((k 0 (+ k 1)))
		  ((= k per-square))
		(make (+ (/ i cols) (/ (random 1.) cols))
		      (+ (/ j rows) (/ (random 1.) rows))))))

	  (do ()
	      ((= i n))
	    (make (random 1.) (random 1.)))
		      
	  (values xs ys))))))

#|
(define (show-points xs ys)
  (let ((g (make-graphics-device)))
    (graphics-operation g 'set-foreground-color "white")
    (graphics-operation g 'set-background-color "black")
    (graphics-set-coordinate-limits g -.01 -.01 1.01 1.01)
    (graphics-enable-buffering g)
    (graphics-clear g)

    (graphics-operation g 'set-foreground-color "cyan")
    (do ((i 0 (+ i 1)))
	((= i (vector-length xs)))
      (graphics-operation g 'fill-circle (vector-ref xs i) (vector-ref ys i)
			  0.007))
    (graphics-flush g)
    g))
|#