(declare (usual-integrations))

(load-option 'hash-table)

(define *STORE-SIZE* 20)

(define-structure
  (simulation
   (conc-name simulation.)
   (constructor %make-simulation ())
   (print-procedure
    (standard-unparser-method 'SIMULATION
      (lambda (simulation port)
	(write-char #\space port)
	(write (hash (simulation.environment simulation)) port)))))

  ;; A simulation has a MAKE-SCHEDULER which is used to set up the simulation.
  ;; This returns a scheduler which is placed in the SCHEDULER-RESUME slot.
  ;; The SCHEDULER-RESUME should continue running the simulation.

  (make-scheduler simulation.default-scheduler)
  (scheduler-resume #F)

  (agenda #F)				; Simulation Agenda

  (n-processors 0)			; number of processors
  (processors (make-vector 10 #F))		; A vector of processor stores.

  ;; a vector of neighbour IDs, locations, neighbours etc
  (neighbours)
  (x) (y) (z)

  ;; ENVIRONMENT is a Scheme environment into which the code is loaded.
  ;;   This environment is shared between all the processors.
  ;; LINKER-MAP map of names to slots & default values (descriptors)
  ;; CACHES is a weak list of caches so we can nullify them
  ;; STORE-LENGTH is the length of the vector allocated to each processor
  ;; NEXT-FREE-SLOT is used for allocating store slots to global variables.

  (environment (let () (the-environment)) read-only true)
  (linker-map (make-eq-hash-table) read-only true)
  (caches '())
  (store-length (+ store:first-free *store-size*))
  (next-free-slot store:first-free)

  ;; Initial program expression, built from load files.
  (initial-program
   (lambda (cont fuel store)
     cont fuel store			; ignored
     (error "No program loaded into simulation")))

  (interrupt/1 #F)			; used by interrupt-kind/1
  (interrupt/2 #F)			; used by interrupt-kind/2
  (interrupt/3 #F)			; used by interrupt-kind/3

  ;; Other IO devices

  (io-devices  (make-eq-hash-table) read-only true)

  ;; Auxilliary stuff
  (display  #F)                         ; the window or whatever
  (display-states #F)			; a vector of processor `widgets'
  (extensions (make-eq-hash-table))
)

(define (simulation.for-each-processor simulation receiver)
  (let ((n (simulation.n-processors simulation))
	(v (simulation.processors simulation)))
    (do ((i 0 (+ i 1)))
	((= i n))
      (receiver (vector-ref v i)))))

;; Extensions to a simulation

(define (simulation.define sim name value)
  (hash-table/put! (simulation.extensions sim) name value))

(define (simulation.set! sim name value)
  (hash-table/put! (simulation.extensions sim) name value))

(define (simulation.get sim name #!optional default)
  (let ((object
	 (hash-table/get (simulation.extensions sim) name
			 (if (default-object? default)
			     simulation.get.bad-marker
			     default))))
    (if (eq? simulation.get.bad-marker object)
	(error "Simulation has no extension slot or method named" name sim)
	object)))

(define (simulation.operation sim name . arguments)
  (apply (simulation.get sim name) sim arguments))

(define simulation.get.bad-marker (list "Absent"))


;; IO devices, i.e. sensors and actuators.  See devices.scm for more
;; details

(define (simulation.define-device sim name device)
  (hash-table/put! (simulation.io-devices sim)
		   name
		   device))

(define (simulation.find-io-device sim name)
  (or (hash-table/get (simulation.io-devices sim) name #F)
      (error "Simulation has no sensor or actuator with name " name sim)))

(define (simulation.initialize-io-devices sim)
  (hash-table/for-each (simulation.io-devices sim)
    (lambda (name device)
      name ; ignored
      (io-device.initialize! device sim))))

;; Simulation events are thunks (different from gunk processor thread
;; events below).

(define (simulation.enqueue sim time event)
  (agenda.add (simulation.agenda sim) time event))

(define (simulation.enqueue+ sim time event)
  (let ((agenda (simulation.agenda sim)))
    (agenda.add agenda (+ time (agenda.time agenda)) event)))

;; Threads
;;
;; Each gunk processor is multi-threaded.  The current running thread is
;; accessible from the THREAD slot of the processor's store.
;;
;; A thread is in one of five states:
;;  NEW - just created, never run
;;  RUNNABLE - waiting for time slice
;;  RUNNING  -
;;  BLOCKED  - waiting for an event
;;  DEAD     - stopped for the last time.
;;
;; Each processor has two slots for the threads

(define-structure
    (thread
     (conc-name thread.)
     (constructor %make-thread (processor))
     (print-procedure
      (standard-unparser-method 'GUNK-THREAD
	(lambda (thread port)
	  (if (thread.processor thread)
	      (begin
		(write-char #\space port)
		(write `(proc ,(processor.index (thread.processor thread)))
		       port)
		(write-char #\space port)
		(write (processor.simulation (thread.processor thread)) port)))
	  (write-char #\space port)
	  (write (thread.state thread) port)))))

  (processor #F read-only true)		; on which thread lives

  ;; Ready threads are kept in a doubly linked circular list.
  (next #F)
  (prev #F)

  ;; STATE is either 'NEW, a vector of EVENTS (when blocked)
  ;; 'RUNNABLE if runnable OR running, 'DEAD if dead.
  (state 'NEW)

  ;; The TERMINATION-EVENT is signalled when the thread becomes dead.
  ;; It contains the final (or exit) value computed by the thread.
  (termination-event (%%make-event) read-only true)

  context-method			; How to resume thread...
  context/1				; ... and thread's saved state
  context/2				;
  context/3				;
  context/4				;
  context/5				;
  context/6				;
  )

(define-integrable (thread.dead? thread)
  (eq? (thread.state thread) 'DEAD))

(define-integrable (thread.blocked? thread)
  (vector? (thread.state thread)))

(define-integrable (thread.new? thread)
  (eq? (thread.state thread) 'NEW))

(define-integrable (thread.runnable? thread)
  (eq? (thread.state thread) 'RUNNABLE))


(define (thread.make-runnable! thread)
  ;; Assumes that thread is not already enqueued.
  (guarantee-thread thread 'thread.make-runnable!)
  (if (thread.dead? thread)
      (error "Gunk thread is dead" thread 'thread.make-runnable!))
  (if (not (thread.runnable? thread)) ; i.e. blocked or 'NEW
      (let ((store (processor.store (thread.processor thread))))
	(if (thread.new? thread)
	    (vector-set! store store:threads
			 (cons thread (vector-ref store store:threads))))
	(set-thread.state! thread 'RUNNABLE)
	(let ((current-thread (vector-ref store store:thread)))
	  (set-thread.prev! thread (thread.prev (thread.next current-thread)))
	  (set-thread.next! thread (thread.next current-thread))
	  (set-thread.prev! (thread.next current-thread) thread)
	  (set-thread.next! current-thread thread))))
  unspecific)


(define (%thread.suspend thread)
  ;; Assume that thread is running.  Unlink it from the running queue.
  (let ((processor (thread.processor thread)))
    (let ((store (processor.store processor)))
      ;; If THREAD is the running thread, make running thread something else.
      (if (eq? thread (vector-ref store store:thread))
	  (if (eq? thread (thread.next thread))
	      (vector-set! store store:thread (make-idle-thread store))
	      (vector-set! store store:thread (thread.next thread))))
      (let ((next (thread.next thread))
	    (prev (thread.prev thread)))
	(set-thread.next! prev next)
	(set-thread.prev! next prev)
	(set-thread.next! thread #F)
	(set-thread.prev! thread #F)))))
	    
(define (make-idle-thread store)
  (let ((processor (store.processor store)))
    (let ((thread (%make-thread processor)))
      (set-thread.context-method! thread resume/idle)
      (thread.make-runnable! thread)
      thread)))
      
(define (resume/idle store fuel)
  (let ((thread (vector-ref store store:thread)))
    (vector-set! store store:thread (thread.next thread))
    ;;(pp `(store: ,store idle ,thread))
    fuel				; use up all the fuel
    -1))

;;(define (idle-thread? thread)
;;  (eq? (thread.context-method thread) resume/idle))

(define (thread.kill! thread value)
  ;; From any state, make thread dead.
  ;; If the thread becomes dead, VALUE is registered as its final value.
  (define (remove-from-threads-list)
    (let ((store (processor.store (thread.processor thread))))
      (vector-set! store store:threads
		   (delq! thread (vector-ref store store:threads)))))
  (guarantee-thread thread 'thread.kill!)
  (cond ((thread.dead? thread))
	((thread.new? thread))
	((thread.blocked? thread)
	 (thread.unregister! thread)
	 (remove-from-threads-list))
	((thread.runnable? thread)
	 (%thread.suspend thread)
	 (remove-from-threads-list)))
  (if (not (event.signalled? (thread.termination-event thread)))
      (event.signal! (thread.termination-event thread) value))
  (set-thread.state! thread 'DEAD))



;; EVENTs are the things a thread blocks on.
;; A thread blocks on a set of events.
;; When an event is signalled, all waiting threads unblock and return
;; from the SELECT call.

(define-structure
    (event
     (conc-name event.)
     (constructor %%make-event ()))

  (blocked-threads  '())		; thread waiting for this event
  (value  event-pending-signal)
  )

(define event-pending-signal (list 'event-pending-signal))

(define (event.clear! event)
  (guarantee-event event 'EVENT.CLEAR!)
  (set-event.value! event event-pending-signal))

(define (event.signalled? event)
  (guarantee-event event 'EVENT.SIGNALLED?)
  (not (eq? (event.value event) event-pending-signal)))

(define (event.signal! event value)
  ;; Signal an event.
  ;; All threads waiting on the event are currently blocked.
  ;; Unblock all threads waiting on that event, making them RUNNABLE.
  (define (unblock thread)
    (let ((v (thread.state thread)))
      ;; Find the index of the event
      (let ((index
	     (or (vector-find-next-element v event)
		 (error "Problem in EVENT.SIGNAL!" event thread))))
	;; Unregister thread from all events it is waiting on
	(thread.unregister! thread)
	;; Set up return value.
	(set-thread.context/2! thread (cons index value))
	(thread.make-runnable! thread))))

  (guarantee-event event 'EVENT.SIGNAL!)
  (set-event.value! event value)
  (for-each unblock (event.blocked-threads event)))


(define (thread.unregister! thread)
  ;; Pre: THREAD is blocked.
  ;; Unregister thread from all events it is waiting.
  ;; Thread is left in indeterminate state.
  (for-each-vector-element (thread.state thread)
    (lambda (event)
      (and event
	   (let ((remaining-blocked-threads
		  (delq! thread (event.blocked-threads event))))
	     (set-event.blocked-threads! event remaining-blocked-threads))))))
  


(define (%select cont fuel store events)
  ;; CONT is passed a pair (index . value), or #F
  ;; EVENTS is a vector of event structures
  (define %select-cost 5)
  (define (->event e)
    (cond ((event? e) e)
	  ((false? e) #F)
	  ((thread? e) (thread.termination-event e))
	  (else (error:wrong-type-argument e "event or thread" 'select))))
  (define (false-or-event.signalled? object)
    (and object
	 (event.signalled? object)))
  (let ((events (vector-map events ->event))
	(fuel*  (fix:- fuel %select-cost)))
    (cond ((not (vector-find-next events identity-procedure))
	   ;; I.e. all are false
	   (cont #F fuel* store))
	  ;; If one of the events has already been signalled, return that event
	  ((vector-find-next events false-or-event.signalled?)
	   => (lambda (i)
		(let ((value (event.value (vector-ref events i))))
		  ;; i, value
		  (cont (cons i value) 
			fuel*
			store))))
	  (else
	   ;; ... Otherwise add thread to wait lists of events ...
	   (let ((current-thread (vector-ref store store:thread)))

	     (for-each-vector-element events
	       (lambda (event)
		 (if event
		     (set-event.blocked-threads!
		      event
		      (cons current-thread
			    (event.blocked-threads event))))))
	     ;; ... suspend the thread ...
	     (set-thread.state! current-thread events)
	     (%thread.suspend current-thread)

	     ;; ... store context ...
	     (set-thread.context-method! current-thread resume/r)
	     (set-thread.context/1!      current-thread cont)
	     ;;(set-thread.context/2!    current-thread 'FILL-IN-LATER))

	     ;; .. and finish this timeslice and processor `step'.
	     fuel*)))))


(define (%make-timeout-event cont fuel store delay)
  (let ((sim   (store.simulation store))
	(event (%%make-event))
	(now   (- (vector-ref store store:counter) fuel)))
    (if (> delay 0)
	(simulation.enqueue sim
			    (+ delay now)
			    (named-lambda (timeout)
			      (event.signal! event 'timeout)))
	(set-event.value! event now))
    (cont event (fix:- fuel 1) store)))
      
(define (%make-event cont fuel store)
  (cont (%%make-event) (fix:- fuel 1) store))
      


(define (vector-find-next vector predicate?)
  (guarantee-vector vector 'vector-find-next)
  (%subvector-find-next vector 0 (vector-length vector) predicate?))

(define (subvector-find-next vector start end predicate?)
  (guarantee-subvector vector start end 'subvector-find-next)
  (%subvector-find-next vector start end predicate?))

(define (%subvector-find-next vector start end predicate?)
  (let loop ((i start))
    (cond ((fix:= i end) #F)
	  ((predicate? (vector-ref vector i)) i)
	  (else (loop (+ i 1))))))

;; The STORE
;;
;; The store represents all the state of a processor, including the values
;; in the top-level environment.  We use in in lieu of a processor
;; object.  The store is a vector.
;;   - Slot 0 is a tag
;;   - Slot 1 is the simulation
;;   - Slot 2 is the index within the simulation
;; Slot 1 enables the linker (cache manager) to access the rest of the
;; processor state, including the simulation's environment.  Other
;; slots are used for machine-visible purposes, for example, saving
;; the context so that we can resume the processor later.
;;
;; The linker can do with the rest of the vector (slots
;; store:first-free and above) whatever it likes.  Note that the
;; linker is also allowed to replace the store with another, larger,
;; store, provided it copies all the existing slots.

;; Slot uses:
(define-integrable store:tag 0); 
(define-integrable store:simulation 1)
(define-integrable store:index 2)
(define-integrable store:counter 3)
(define-integrable store:thread 4)      ; current running thread
(define-integrable store:threads 5)      ; all running, runnable or blocked threads.

(define-integrable store:comms 7)	; for use by comms simulation
(define-integrable store:first-free 8)

(define-integrable store-tag (string->symbol "#[processor]"))

(define store-record-type
  (named-structure/set-tag-description!
   store-tag
   ((access make-define-structure-type system-global-environment)
    'vector
    "processor"
    '(          simulation       index       thread)
    (list store:simulation store:index store:thread)
    ((access standard-unparser-method system-global-environment) 'store #f))))

(define (store? object)
  (and (vector? object)
       (> (vector-length object) 0)
       (eq? (vector-ref object 0) store-tag)))

(define-integrable (store.simulation store) (vector-ref store store:simulation))
(define-integrable (store.index store)      (vector-ref store store:index))

(define-integrable (store.processor store) store)
(define-integrable (processor.store processor) processor)
(define-integrable (processor.simulation processor)
  (store.simulation (processor.store processor)))
(define-integrable (processor.index processor)
  (store.index (processor.store processor)))


(define (make-store simulation index)
  (let ((store  (make-vector (simulation.store-length simulation) #F)))
    (vector-set! store store:tag store-tag)
    (vector-set! store store:simulation simulation)
    (vector-set! store store:index index)
    (vector-set! store store:counter 0)
    ;;(vector-set! store store:context-method resume/error-initialized)
    (vector-set! store store:thread #F)
    (vector-set! store store:comms (initial-comms-slot-value index))
    (hash-table/for-each (simulation.linker-map simulation)
      (lambda (name descriptor)
	name				; ignored
	(if (and (descriptor.slot descriptor)
		 (descriptor.default-value? descriptor))
	    (vector-set! store
			 (descriptor.slot descriptor)
			 (descriptor.default-value descriptor)))))
    store))

(define (initial-comms-slot-value index)
  index ; ignored
  #F)

(define (resume/error-initialized store fuel)
  fuel					; ignored
  (error (string-append "Simulation has no program loaded.  (At processor #"
			(write-to-string (vector-ref store store:index))
			")")
	 (store.simulation store)))

(define (store.saved-context store)
  (error "NOT THREAD")
  (subvector store store:context-method (+ store:context-last 1)))

(define (store.restore-context! store context)
  (error "NOT THREAD")
  (subvector-move-right! context 0 (vector-length context)
			 store store:context-method))

(define (simulation.add-processor simulation #!optional how-many)
  (let ((how-many (if (default-object? how-many) 1 how-many)))
    (do ((i 0 (+ i 1)))
	((= i how-many)
	 (simulation.reset! simulation)
	 (simulation.n-processors simulation))
      (let ((n (simulation.n-processors simulation))
	    (v (simulation.processors simulation)))
	(if (= n (vector-length v))
	    (set-simulation.processors! simulation (vector-grow v (* n 2))))
	(let ((store (make-store simulation n)))
	  (vector-set! (simulation.processors simulation) n store)
	  (set-simulation.n-processors! simulation (+ n 1)))))))

(define-integrable (cache.name cache) (vector-ref cache 0))
(define-integrable (cache.slot cache) (vector-ref cache 1))
(define-integrable (cache.extra cache) (vector-ref cache 2))
(define-integrable (set-cache.slot! cache slot) (vector-set! cache 1 slot))
(define-integrable (set-cache.extra! cache slot) (vector-set! cache 2 slot))

(define (make-cache name) (vector name #F #F))

(define (clear-cache! cache)
  (set-cache.slot! cache #F)
  (set-cache.extra! cache #F))

(define-integrable (simulation.remember-cache simulation cache)
  (set-simulation.caches! simulation
			  (weak-cons cache (simulation.caches simulation))))

(define-structure
    (descriptor
     (conc-name descriptor.)
     (constructor make-descriptor (name)))
  (name #F read-only #T)
  (slot #F)
  (default-value? #F)
  (default-value #F)			; value if DEFAULT-VALUE? is #T
  (files #F))

(define (simulation.linker-descriptor sim name)
  (or (hash-table/get (simulation.linker-map sim) name #F)
      (let ((descriptor (make-descriptor name)))
	(hash-table/put! (simulation.linker-map sim) name descriptor)
	descriptor)))

(define (global-ref/fault store cache)
  (let ((sim (store.simulation store)))
    (let ((descriptor (simulation.linker-descriptor sim (cache.name cache))))
      (simulation.remember-cache sim cache)
      (let ((slot (descriptor.slot descriptor)))
	(if slot
	    (begin
	      (set-cache.slot! cache slot)
	      (global-ref store cache))
	    (begin
	      (allocate-store-slot/ref-or-set! cache sim descriptor)
	      (global-ref store cache)))))))

(define (global-set!/fault store cache value)
  (let ((sim (store.simulation store)))
    (let ((descriptor (simulation.linker-descriptor sim (cache.name cache))))
      (simulation.remember-cache sim cache)
      (let ((slot (descriptor.slot descriptor)))
	(if slot
	    (begin
	      (set-cache.slot! cache slot)
	      (global-set! store cache value))
	    (begin
	      (allocate-store-slot/ref-or-set! cache sim descriptor)
	      (global-set! store cache value)))))))

#|
;; This version works just like SET!

(define (global-define!/fault store cache value)
  (let ((sim (store.simulation store)))
    (let ((descriptor (simulation.linker-descriptor sim (cache.name cache))))
      (simulation.remember-cache sim cache)
      (let ((slot (descriptor.slot descriptor)))
	(if slot
	    (begin
	      (set-cache.slot! cache slot)
	      (global-define! store cache value))
	    (begin
	      (allocate-store-slot/ref-or-set! cache sim descriptor)
	      (global-define! store cache value)))))))
|#

(define (global-define!/fault store cache value)
  ;; cache is (name #F #F) or
  ;;          (name #F value*)  where (not (eq? value value*))
  (let ((sim (store.simulation store)))
    (let ((descriptor (simulation.linker-descriptor sim (cache.name cache))))
      (simulation.remember-cache sim cache)
      (let ((slot (descriptor.slot descriptor)))
	(if slot
	    (begin
	      (set-cache.slot! cache slot)
	      (global-define! store cache value))
	    (let ((name  (cache.name cache)))
	      (if (not (descriptor.default-value? descriptor))
		  (let ((default-value (find-default-definition sim name)))
		    (set-descriptor.default-value?! descriptor #T)
		    (set-descriptor.default-value! descriptor default-value)))
	      (let ((default-value (descriptor.default-value descriptor)))
		(set-cache.extra! cache default-value)
		(if (and (eq? value default-value)
			 (memq name (environment-bound-names
				     (simulation.environment sim))))
		    (begin
		      (global-define! store cache value))
		    (begin
		      (allocate-store-slot/ref-or-set! cache sim descriptor)
		      (global-define! store cache value))))))))))


(define (allocate-store-slot/ref-or-set! cache simulation descriptor)
  (let* ((name  (cache.name cache))
	 (default-value (find-default-definition simulation name)))
    (allocate-store-slot name simulation default-value descriptor)
    (set-cache.slot! cache (descriptor.slot descriptor))))

(define (allocate-store-slot name simulation value descriptor)
  (let ((slot       (+ (simulation.next-free-slot simulation) 1))
	(slot-limit (simulation.store-length simulation)))
    (if (>= slot slot-limit)
	(error "Too few store slots.  Set *STORE-SIZE* to at least"
	       slot-limit
	       (error-irritant/noise " and re-create simulation. ")
	       name))
    (set-simulation.next-free-slot! simulation slot)
    (set-descriptor.slot! descriptor slot)
    (set-descriptor.default-value?! descriptor #T)
    (set-descriptor.default-value!  descriptor value)
    (simulation.for-each-processor simulation
      (lambda (processor)
	(vector-set! (processor.store processor) slot value)))))

(define (find-default-definition simulation name)
  (let ((defaults-environment (simulation.environment simulation)))
    (if (environment-bound? defaults-environment name)
	(let ((global-value (environment-lookup defaults-environment name))
	      (gunk-bound?
	       (memq name (environment-bound-names defaults-environment))))
	  (if (procedure? global-value)
	      (if gunk-bound?
		  global-value
		  (begin
		    (warn "Using non-gunk version of procedure:" name)
		    (convert-normal-procedure global-value)))
	      (begin
		(if (not gunk-bound?)
		    (warn "Using non-gunk value of variable:" name))
		global-value)))
	(string->uninterned-symbol
	 (string-append "#[unassigned " (symbol-name name) "]")))))

(define (convert-normal-procedure proc)
  (define (default)
    (lambda (k f s . args)
      (k (apply proc args) f s)))
  (let ((arity (procedure-arity proc)))
    (if (eq? (car arity) (cdr arity))
	(case (car arity)
	  ((0)  (lambda (k f s) (k (proc) f s)))
	  ((1)  (lambda (k f s a) (k (proc a) f s)))
	  ((2)  (lambda (k f s a b) (k (proc a b) f s)))
	  ((3)  (lambda (k f s a b c) (k (proc a b c) f s)))
	  ((4)  (lambda (k f s a b c d) (k (proc a b c d) f s)))
	  (else (default)))
	(default))))



(define *file-global-references*)

(define (current-simulation.notice-file-definitions . args)
  args					; ignored
  (warn "File transformed with obsolete version of CPS. Re-CPS file:"
	(enough-namestring (current-load-pathname))))

(define (current-simulation.notice-globals
	 defined-names assigned-names referenced-names)
  (set! *file-global-references*
	(cons (vector (current-load-pathname)
		      defined-names assigned-names referenced-names)
	      *file-global-references*)))

(define (check-file-global-references sim file-global-references)
  ;; Sanity check the global references in the user's files
  (define (file.pathname v) (vector-ref v 0))
  (define (file.defs v) (vector-ref v 1))
  (define (file.sets v) (vector-ref v 2))
  (define (file.refs v) (vector-ref v 3))

  (define (ucons elt lst)
    (if (memq elt lst)
	lst
	(cons elt lst)))

  (let ((names (make-eq-hash-table))
	(defs  (make-eq-hash-table))
	(sets  (make-eq-hash-table))
	(refs  (make-eq-hash-table)))

    (define (process-file file-info)
      (define (adder set)
	(lambda (name)
	  (hash-table/put! set name 
			   (ucons (file.pathname file-info)
				  (hash-table/get defs name '())))))
      
      (for-each (adder defs) (file.defs file-info))
      (for-each (adder sets) (file.sets file-info))
      (for-each (adder refs) (file.refs file-info))
      (for-each (adder names)
	(append (file.defs file-info) (file.sets file-info)
		(file.refs file-info))))
    
    (define (files-warning name.files)
      (apply warn "  " (car name.files)
	     (error-irritant/noise "  in ")
	     (map enough-namestring (cdr name.files))))

    (for-each process-file file-global-references)

    (let ((duplicate-definitions
	   (list-transform-positive (hash-table->alist defs)
	     (lambda (name.files)
	       (> (length (cdr name.files)) 1)))))
      (if (pair? duplicate-definitions)
	  (begin
	    (warn "Duplicate definitions:")
	    (for-each files-warning duplicate-definitions))))

    (let ((undefined-references-no-default
	   (list-transform-positive (hash-table->alist names)
	     (lambda (name.files)
	       (let ((name (car name.files)))
		 (and (not (hash-table/get defs name #F))
		      (not (environment-bound? (simulation.environment sim)
					       name))))))))
      (if (pair? undefined-references-no-default)
	  (begin
	    (warn "References to bindings that are never defined and have no default:")
	    (for-each files-warning undefined-references-no-default))))

    (let ((minimum-store-size (hash-table/count names)))
      (if (< *STORE-SIZE* minimum-store-size)
	  (begin
	    (warn "*STORE-SIZE*" (list *STORE-SIZE*)
		  (error-irritant/noise " is probably too small."))
	    (warn "Set *STORE-SIZE* to at least"
		  minimum-store-size
		  (error-irritant/noise " and re-create simulation. ")))))

    (cond ((hash-table/get names 'MESSAGE-INTERRUPT-HANDLER #F)
	   => (lambda (name.files)
		(warn "Using MESSAGE-INTERRUPT-HANDLER is obsolete.")
		(files-warning (cons 'MESSAGE-INTERRUPT-HANDLER name.files))))
	  )))


(define (simulation.load sim . programs)
  ;; PROGRAMS are pathnames, strings or the symbols 'INIT and 'LIB
  (fluid-let ((*simulation* sim)
	      (*file-global-references* '()))
    (define (deref-filename file)
      (cond ((eq? file 'init) (merge-pathnames "init" hlsim-load-directory))
	    ((eq? file 'lib)  (merge-pathnames "lib"  hlsim-load-directory))
	    (else file)))
    (define (load-1 file)
      (let ((file (deref-filename file)))
	(let ((expression (load file (simulation.environment sim)))
	      (old-initial-program (simulation.initial-program sim)))
	  (let ((new-initial-program
		 (lambda (cont fuel store)
		   (old-initial-program
		    (lambda (value fuel store)
		      value		; ignored
		      (expression cont fuel store))
		    fuel store))))
	    (set-simulation.initial-program! sim new-initial-program)))))
    (simulation.linker.reset! sim)
    (set-simulation.initial-program! sim null-program)
    (let load-many ((programs programs))
      (if (pair? programs)
	  (for-each load-many programs)
	  (load-1 programs)))
    (check-file-global-references sim *file-global-references*)
    (simulation.reset! sim)))

(define (simulation.reset! sim)
  (guarantee-simulation sim 'SIMULATION.RESET!)
  ;; Reset simulation as if never run.
  (set-simulation.interrupt/1! sim #F)
  (set-simulation.interrupt/2! sim #F)
  (set-simulation.interrupt/3! sim #F)
  (for-each clear-cache! (weak-list->list (simulation.caches sim)))
  (set-simulation.caches! sim '())
  (set-simulation.scheduler-resume! sim #F))

(define (simulation.linker.reset! sim)
  ;; Reinitialize linker
  (hash-table/clear! (simulation.linker-map sim))
  (set-simulation.next-free-slot! sim store:first-free))

(define (null-program cont fuel store)
  (cont unspecific fuel store))

(define *simulation*)
(define simulation.default-scheduler)

(define (simulation.run sim . options)
  ;; Resume execution if already running, otherwise start from scratch.
  (guarantee-simulation sim 'SIMULATION.RUN)
  (fluid-let ((*simulation* sim))
    (if (simulation.scheduler-resume sim)
	((simulation.scheduler-resume sim))
	(let ((scheduler (apply (simulation.make-scheduler sim) sim options)))
	  (set-simulation.scheduler-resume! sim scheduler)
	  (simulation.initialize-io-devices sim)
	  (scheduler)))))


(define (simulation.initialize-processor-programs! sim continuation:done)
  ;; Set up processors to run the program and return to CONTINUATION:DONE
  ;; when the whole program has finished.

  (define (initialize-processor! processor)
    (let ((store (processor.store (processor.store processor))))
      (let ((initial-thread (%make-thread processor)))
	(vector-set! store store:counter 0)
	(vector-set! store store:thread initial-thread)
	(vector-set! store store:threads (list initial-thread))
	(set-thread.next! initial-thread initial-thread)
	(set-thread.prev! initial-thread initial-thread)
	(set-thread.state! initial-thread 'RUNNABLE)
	(set-thread.context-method! initial-thread resume/0)
	(set-thread.context/1! initial-thread (simulation.initial-program sim))
	(set-thread.context/2! initial-thread continuation:done))))

  (simulation.for-each-processor sim initialize-processor!))

(define (processor.termination-notice processor fuel value)
  (define (fmt number width)
    (string-pad-left (number->string number) width))
  (let ((n (simulation.n-processors (processor.simulation processor))))
    (newline)
    (write-string "; processor")
    (let ((digits (+ (truncate->exact (/ (log n) (log 10))) 2)))
      (write-string (fmt (processor.index processor) digits)))
    (write-string " finished fuel")
    (write-string (fmt fuel 3))
    (write-string "  ")
    (write value)))

(define (transform-user-expression code simulation)
  ;; User code can be a scheme thunk, pre-transformed gunk thunk, or
  ;; s-expression.
  (define (eval* code)
    (eval code (simulation.environment simulation)))
  (cond ((and (procedure? code) (equal? (procedure-arity code) '(3 . 3)))
	 code)
	((and (procedure? code) (equal? (procedure-arity code) '(0 . 0)))
	 (let ((expression (eval* (rewrite/top-level (procedure-lambda code)))))
	   (lambda (k fuel store)
	     (expression (lambda (thunk fuel store) (thunk k fuel store))
			 fuel store))))
	((procedure? code)
	 (error "Wrong type of procedure:" code))
	(else				; source code
	 (eval* (rewrite/top-level (syntax code))))))

(define (processor.eval processor code)
  (let ((code
	 (transform-user-expression code (processor.simulation processor))))
    (processor.run-expression processor code)))

(define (simulation.eval simulation code #!optional selection)
  selection				; currently ignored
  (guarantee-simulation simulation 'SIMULATION.EVAL)
  (let ((code (transform-user-expression code simulation))
	(processors (simulation.processors simulation)))
    (make-initialized-vector (simulation.n-processors simulation)
      (lambda (i)
	(processor.run-expression (vector-ref processors i) code)))))

(define (processor.run-expression processor gunk-thunk)
  (let* ((fin? #F)
	 (result #F)
	 (store (processor.store processor))
	 (saved-current-thread (vector-ref store store:thread)))
    (define (continuation:finish! value fuel store)
      store				; ignored
      (set! fin? #T)
      (set! result value)
      fuel)
    (let ((new-thread (%make-thread processor)))
      (vector-set! store store:thread new-thread)
      (set-thread.next! new-thread new-thread)
      (set-thread.prev! new-thread new-thread)
      (set-thread.context-method! new-thread resume/0)
      (set-thread.context/1! new-thread gunk-thunk)
      (set-thread.context/2! new-thread continuation:finish!))
    (let loop ()
      (resume-processor processor 1000)
      (if fin?
	  result
	  (loop)))

    (vector-set! store store:thread saved-current-thread)
    result))


(define-integrable (guarantee-simulation object procedure)
  (if (not (simulation? object))
      (error:wrong-type-argument object "simulation" procedure)))

(define-integrable (guarantee-thread object procedure)
  (if (not (thread? object))
      (error:wrong-type-argument object "gunk thread" procedure)))

(define-integrable (guarantee-event object procedure)
  (if (not (event? object))
      (error:wrong-type-argument object "gunk event" procedure)))
